import imageio.v3 as iio
from pathlib import Path
import numpy as np
from shapely import Polygon
import roifile
from abc import ABC, abstractmethod
from typing import List, Dict, Tuple, Union, Optional, Any

#################################
###### Source Data Handler ######
#################################

class Data(ABC):

    @abstractmethod
    def inject_loaded_data(self, loaded_data: Any) -> None:
        pass

    def __init__(self, filepath: Path) -> None:
        self.filepath = filepath


class Recording(Data):

    def inject_loaded_data(self, loaded_data: np.ndarray) -> None:
        self.zstack = loaded_data
        self.estimated_bit_depth = self._estimate_bit_depth()
        self.preview = self._create_brightness_and_contrast_enhanced_preview()


    def _estimate_bit_depth(self) -> int:
        max_bit_value = self.zstack.max()
        if max_bit_value <= 255:
            estimated_bit_depth = 255
        elif max_bit_value <= 4095:
            estimated_bit_depth = 4095
        elif max_bit_value <= 65535:
            estimated_bit_depth = 65535
        else:
            raise ValueError(f'Max bit value in recording found to be {max_bit_value}, but NA3 currently only handles up to 16-bit recordings!')
        return estimated_bit_depth


    def _create_brightness_and_contrast_enhanced_preview(self, percentile_for_adjustment: int=1) -> np.ndarray:
        raw_image = self.zstack[0, :, :, :].copy() # ensure that dimensions are the same as for ".get_single_frame_as_preview()"
        lower_percentile_bit_value = np.percentile(raw_image, percentile_for_adjustment)
        upper_percentile_bit_value = np.percentile(raw_image, 100-percentile_for_adjustment)
        contrast_adjustment_factor = self.estimated_bit_depth / (upper_percentile_bit_value - lower_percentile_bit_value)
        brightness_adjustment_factor = -(contrast_adjustment_factor * lower_percentile_bit_value)
        raw_image_clipped_at_percentile_bit_values = self._clip_image_at_bit_values(raw_image, lower_percentile_bit_value, upper_percentile_bit_value)
        brightness_contrast_adjusted_image = contrast_adjustment_factor * raw_image_clipped_at_percentile_bit_values + brightness_adjustment_factor
        return brightness_contrast_adjusted_image
        

    def _compute_contrast_and_brightness_adjustment_factors(self, raw_image: np.ndarray, percentile_for_adjustment: int=1) -> Tuple[float, float]:
        lower_percentile_bit_value = np.percentile(raw_image, percentile_for_adjustment)
        upper_percentile_bit_value = np.percentile(raw_image, 100-percentile_for_adjustment)
        contrast_adjustment = self.estimated_bit_depth / (upper_percentile_bit_value - lower_percentile_bit_value)
        brightness_adjustment = -(contrast_adjustment * lower_percentile_bit_value)
        return contrast_adjustment, brightness_adjustment


    def _clip_image_at_bit_values(self, raw_image: np.ndarray, min_bit_value: float, max_bit_value: float) -> np.ndarray:
        raw_image[raw_image <= min_bit_value] = min_bit_value
        raw_image[raw_image >= max_bit_value] = max_bit_value
        return raw_image


class ROI(Data):

    def inject_loaded_data(self, loaded_data: List[Tuple[int, int]]) -> None:
        self.boundary_coords = loaded_data
        self.as_polygon = self._convert_to_valid_polygon()


    def _convert_to_valid_polygon(self) -> Polygon:
        roi_as_polygon = Polygon(self.boundary_coords)
        assert roi_as_polygon.is_valid, f'Something went wrong when trying to create a Polygon out of your ROI: {self.filepath}.'
        return roi_as_polygon



################################
###### Source Data Loader ######
################################


class DataLoader(ABC):

    @abstractmethod
    def load_data(self) -> Union[Data, Tuple[Data, Data]]:
        # This method will be called when the data should be loaded for analysis
        pass

    def __init__(self, filepath: Path) -> None:
        self.filepath = filepath



class RecordingLoader(DataLoader):


    @abstractmethod
    def _get_all_frames(self) -> np.ndarray: 
        # To be implemented in individual subclasses
        # Shape of returned numpy array: [frames, rows, cols, color_channels]
        pass


    def _load_all_frames(self) -> np.ndarray: 
        all_frames = self._get_all_frames()
        all_frames = self._validate_shape_and_convert_to_grayscale_if_possible(all_frames)
        return all_frames


    def load_data(self) -> Recording:
        all_frames = self._load_all_frames()
        recording = Recording(self.filepath)
        recording.inject_loaded_data(all_frames)
        return recording


    def _validate_shape_and_convert_to_grayscale_if_possible(self, zstack: np.ndarray) -> np.ndarray:
        self._validate_correct_array_shape(zstack)
        if zstack.shape[3] > 1:
            if self._check_if_color_channels_are_redunant(zstack) == True:
                zstack = self._convert_to_grayscale(zstack)
        return zstack


    def _validate_correct_array_shape(self, zstack: np.ndarray) -> None:
        assert len(zstack.shape) == 4, ('The shape of the zstack numpy array is not correct. It should be a 4 dimensional array, like '
                                        f'[frames, rows, cols, color channels]. However, the current shape is: {zstack.shape}.')
        assert zstack.shape[3] in [1, 3], ('The color channels of the recording you attempted to load are incorrect. Currently, only single '
                                           f'channel or RGB (i.e. 1 or 3 color channels) are supported. However, your data has: {zstack.shape[3]}.')
        
    
    def _check_if_color_channels_are_redunant(self, zstack: np.ndarray) -> bool:
        reference_channel_idx = 0
        color_channels_are_equal = []
        for idx_of_channel_to_compare in range(1, zstack.shape[3]):
            if np.array_equal(zstack[:, :, :, reference_channel_idx], zstack[: , :, :, idx_of_channel_to_compare]) == True:
                color_channels_are_equal.append(True)
            else:
                color_channels_are_equal.append(False)
                break
        return all(color_channels_are_equal)


    def _convert_to_grayscale(self, zstack: np.ndarray) -> np.ndarray:
        return zstack[:, :, :, 0:1]
        
        
class AVILoader(RecordingLoader):

    def _get_all_frames(self) -> np.ndarray: 
        return iio.imread(self.filepath)
        


class ROILoader(DataLoader):

    @abstractmethod
    def _get_boundary_coords(self) -> List[Tuple[int, int]]: 
        # To be implemented in individual subclasses
        # Return a list of Tuples, where each tuple represents one boundary point: (row_coord, col_coord)
        pass

    
    def load_data(self) -> ROI: 
        boundary_coords = self._get_boundary_coords()
        boundary_coords = self._add_first_boundary_point_also_add_end_to_close_roi(boundary_coords)
        roi = ROI(self.filepath)
        roi.inject_loaded_data(boundary_coords)
        return roi


    def _add_first_boundary_point_also_add_end_to_close_roi(self, boundary_coords: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
        first_boundary_point_coords = boundary_coords[0]
        boundary_coords.append(first_boundary_point_coords)
        return boundary_coords


class ImageJROILoader(ROILoader):

    def _get_boundary_coords(self) -> List[Tuple[int, int]]:
        roi_file_content = roifile.ImagejRoi.fromfile(self.filepath)
        assert type(roi_file_content) != list, (
            'NeuralActivityCubic can currently only handle loading ROI files that contain exactly one ROI per file. '
            f'However, the ROI file you selected ("{self.filepath}") contains more than a single ROI! Please adjust '
            'the file accordingly and retry.'
        )
        row_coords = roi_file_content.coordinates()[:, 1]
        col_coords = roi_file_content.coordinates()[:, 0]
        boundary_row_col_coords = list(zip(row_coords, col_coords))
        return boundary_row_col_coords


##############################
###### Loader Factories ######
##############################

class DataLoaderFactory(ABC):

    @property
    @abstractmethod
    def supported_extensions_per_data_loader(self) -> Dict[DataLoader, List[str]]:
        pass

    
    @property
    def all_supported_extensions(self) -> List[str]:
        all_supported_extensions = []
        for value in self.supported_extensions_per_data_loader.values():
            all_supported_extensions += value
        return all_supported_extensions

    
    def get_loader(self, filepath: Path) -> DataLoader:
        self._assert_validity_of_filepath(filepath)
        data_loader = self._get_loader_for_file_extension(filepath)
        return data_loader
        

    def _assert_validity_of_filepath(self, filepath: Path) -> None:
        assert isinstance(filepath, Path), f'filepath must be an instance of a pathlib.Path. However, you passed {filepath}, which is of type {type(filepath)}'
        assert filepath.exists(), f'The filepath you provided ({filepath}) does not seem to exist!'
        

    def _get_loader_for_file_extension(self, filepath: Path) -> DataLoader:
        matching_loader = None
        for loader_subclass, supported_extensions in self.supported_extensions_per_data_loader.items():
            if filepath.suffix in supported_extensions:
                matching_loader = loader_subclass(filepath)
                break
        if matching_loader == None:
            raise NotImplementedError('It seems like there is no DataLoader implemented for the specific filetype you´re trying to load - sorry!')
        return matching_loader


class RecordingLoaderFactory(DataLoaderFactory):

    @property
    def supported_extensions_per_data_loader(self) -> Dict[RecordingLoader, List[str]]:
        supported_extensions_per_data_loader = {AVILoader: ['.avi']}
        return supported_extensions_per_data_loader



class ROILoaderFactory(DataLoaderFactory):

    @property
    def supported_extensions_per_data_loader(self) -> Dict[ROILoader, List[str]]:
        supported_extensions_per_data_loader = {ImageJROILoader: ['.roi']}
        return supported_extensions_per_data_loader



def get_filepaths_with_supported_extension_in_dirpath(dirpath: Path, all_supported_extensions: List[str], max_results: Optional[int]=None) -> List[Path]:
    all_filepaths_with_supported_extension = []
    for elem in dirpath.iterdir():
        if elem.is_file() == True:
            if elem.suffix in all_supported_extensions:
                all_filepaths_with_supported_extension.append(elem)
    if type(max_results) == int:
        assert len(all_filepaths_with_supported_extension) <= max_results, (
            f'There are more than {max_results} file(s) of supported type in {dirpath}, '
            f'but only a maximum of {max_results} are allowed. Please remove at least '
            f'{len(all_filepaths_with_supported_extension) - max_results} of the following' 
            f'files and try again: {all_filepaths_with_supported_extension}'
        )
    return all_filepaths_with_supported_extension


######################################
###### Handler for Combinations ######
######################################

class RecLoaderROILoaderCombinator:

        
    def __init__(self, dir_path: Path) -> None:
        self.dir_path = dir_path

    
    def get_all_recording_and_roi_loader_combos(self) -> List[Tuple[RecordingLoader, ROILoader]]:
        recording_loader = self._get_the_recording_loader()
        all_roi_loaders = self._get_all_roi_loaders()
        if len(all_roi_loaders) > 0:
            rec_roi_loader_combos = [(recording_loader, roi_loader) for roi_loader in all_roi_loaders]
        else:
            rec_roi_loader_combos = [(recording_loader, None)]
        return rec_roi_loader_combos
    

    def _get_the_recording_loader(self) -> RecordingLoader:
        recording_loader_factory = RecordingLoaderFactory()
        recording_filepath = get_filepaths_with_supported_extension_in_dirpath(self.dir_path, recording_loader_factory.all_supported_extensions, 1)[0]
        recording_loader = recording_loader_factory.get_loader(recording_filepath)
        return recording_loader


    def _get_all_roi_loaders(self) -> List[ROILoader]:
        roi_loader_factory = ROILoaderFactory()
        all_roi_filepaths = get_filepaths_with_supported_extension_in_dirpath(self.dir_path, roi_loader_factory.all_supported_extensions)
        all_roi_loaders = [roi_loader_factory.get_loader(filepath) for filepath in all_roi_filepaths]
        return all_roi_loaders




import multiprocessing
import numpy as np
import pandas as pd
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes._axes import Axes 
from datetime import datetime, timezone

from typing import Optional, Tuple, Dict, List, Any

from .input import RecordingLoader, Recording, ROILoader, ROI
from . import results
from .analysis import Square


def process_squares(square: Square, configs: Dict[str, Any]) -> Square:
    square.compute_mean_intensity_timeseries(configs['limit_analysis_to_frame_interval'], configs['start_frame_idx'], configs['end_frame_idx'])
    if np.mean(square.mean_intensity_over_time) >= configs['signal_average_threshold']:
        square.detect_peaks(configs['signal_to_noise_ratio'], configs['octaves_ridge_needs_to_spann'], configs['noise_window_size'])
        square.estimate_baseline(configs['baseline_estimation_method'])
        square.compute_area_under_curve()
        square.compute_amplitude_and_delta_f_over_f()
    return square



class AnalysisJob:

    def __init__(self, 
                 number_of_parallel_processes: int,
                 recording_loader: RecordingLoader, 
                 roi_loader: Optional[ROILoader]=None
                ) -> None:
        self.number_of_parallel_processes = number_of_parallel_processes
        self.recording_loader = recording_loader
        self.parent_dir_path = self.recording_loader.filepath.parent
        self.roi_loader = roi_loader
        self.roi_based = (self.roi_loader != None)


    def load_data_into_memory(self) -> None:
        if hasattr(self, 'recording') == False:
            self.recording = self.recording_loader.load_data()
            if self.roi_based == True:
                self.roi = self.roi_loader.load_data()


    def preview_window_size(self, window_size: int) -> Tuple[Figure, Axes]:
        self.load_data_into_memory()
        self._create_squares(window_size)
        if self.roi_based == True:
            roi = self.roi
        else: 
            roi = None
        fig, ax = results.plot_window_size_preview(self.recording.preview, window_size, self.row_cropping_idx, self.col_cropping_idx, roi)
        return fig, ax

    
    def run_analysis(self,
                window_size: int,
                limit_analysis_to_frame_interval: bool,
                start_frame_idx: int,
                end_frame_idx: int,
                signal_average_threshold: float,
                signal_to_noise_ratio: float,
                octaves_ridge_needs_to_spann: float,
                noise_window_size: int,
                baseline_estimation_method: str,
                #include_variance: bool,
                #variance: float
               ) -> None:
        self._set_analysis_start_datetime()
        self.load_data_into_memory()
        self.squares = self._create_squares(window_size)
        configs = locals()
        configs.pop('self')
        with multiprocessing.Pool(processes = self.number_of_parallel_processes) as pool:
            processed_squares = pool.starmap(process_squares, [(square, configs) for square in self.squares])
        self.processed_squares = processed_squares


    def _set_analysis_start_datetime(self) -> None:
            users_local_timezone = datetime.now().astimezone().tzinfo
            self.analysis_start_datetime = datetime.now(users_local_timezone)      


    def _create_squares(self, window_size: int) -> List[Square]:
        self.row_cropping_idx, self.col_cropping_idx = self._get_cropping_indices_to_adjust_for_window_size(window_size)
        upper_left_pixel_idxs_of_squares_in_grid, grid_cell_labels = self._get_positions_for_squares_in_grid(window_size)
        squares = []
        for (upper_left_row_pixel_idx, upper_left_col_pixel_idx), grid_cell_label in zip(upper_left_pixel_idxs_of_squares_in_grid, grid_cell_labels):
            square_row_coords_slice = slice(upper_left_row_pixel_idx, upper_left_row_pixel_idx + window_size)
            square_col_coords_slice = slice(upper_left_col_pixel_idx, upper_left_col_pixel_idx + window_size)
            zstack_within_square = self.recording.zstack[:, square_row_coords_slice, square_col_coords_slice, :]
            squares.append(Square(grid_cell_label, (upper_left_row_pixel_idx, upper_left_col_pixel_idx), zstack_within_square))
        if self.roi_based == True:
            squares_filtered_by_roi = self._filter_squares_based_on_roi(squares)
            squares = squares_filtered_by_roi
        return squares


    def _filter_squares_based_on_roi(self, squares: List[Square]) -> List[Square]:
        filtered_squares = [square for square in squares if square.as_polygon.intersects(self.roi.as_polygon)]
        return filtered_squares

    
    def _get_cropping_indices_to_adjust_for_window_size(self, window_size: int) -> Tuple[int, int]:
        row_cropping_index = (self.recording.preview.shape[0] // window_size) * window_size
        col_cropping_index = (self.recording.preview.shape[1] // window_size) * window_size
        return row_cropping_index, col_cropping_index

    
    def _get_positions_for_squares_in_grid(self, window_size: int) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
        pixel_idxs_of_grid_rows = np.arange(start = 0, stop = self.row_cropping_idx, step = window_size)
        pixel_idxs_of_grid_cols = np.arange(start = 0, stop = self.col_cropping_idx, step = window_size)
        grid_row_labels = np.arange(start = 1, stop = self.row_cropping_idx / window_size + 1, step = 1, dtype = 'int')
        grid_col_labels = np.arange(start = 1, stop = self.col_cropping_idx / window_size + 1, step = 1, dtype = 'int')
        upper_left_pixel_idxs_of_squares_in_grid = []
        grid_cell_labels = []
        for row_pixel_idx, row_label in zip(pixel_idxs_of_grid_rows, grid_row_labels):
            for col_pixel_idx, col_label in zip(pixel_idxs_of_grid_cols, grid_col_labels):
                upper_left_pixel_idxs_of_squares_in_grid.append((row_pixel_idx, col_pixel_idx))
                grid_cell_labels.append((row_label, col_label))
        return upper_left_pixel_idxs_of_squares_in_grid, grid_cell_labels


    def create_results(self, 
                       save_overview_png: bool,
                       save_detailed_results: bool,
                       minimum_activity_counts: int, 
                       window_size: int,
                       signal_average_threshold: float, 
                       signal_to_noise_ratio: float
                      ) -> None:
        self._ensure_results_dir_exists()
        filtered_squares = [square for square in self.processed_squares if square.peaks_count >= minimum_activity_counts]
        roi = self.roi if self.roi_based == True else None
        self.overview_results = results.plot_activity_overview(squares_with_sufficient_activity = filtered_squares, 
                                                               preview_image = self.recording.preview, 
                                                               row_cropping_idx = self.row_cropping_idx, 
                                                               col_cropping_idx = self.col_cropping_idx, 
                                                               window_size = window_size, 
                                                               indicate_activity = True,
                                                               roi = roi)
        if save_overview_png == True:
            self.overview_results[0].savefig(self.results_dir_path.joinpath('overview.png'))
        if save_detailed_results == True:
            self._create_and_save_csv_result_files(filtered_squares)
            self._create_and_save_individual_traces_pdf_result_file(filtered_squares, window_size)
    
        
    def _ensure_results_dir_exists(self) -> None:
        if hasattr(self, 'results_dir_path') == False:
            prefix_with_datetime = self.analysis_start_datetime.strftime('%Y_%m_%d_%H-%M-%S_results_for')
            recording_filename_without_extension = self.recording.filepath.name.replace(self.recording.filepath.suffix, '')
            if self.roi_based == False:
                results_dir_name = f'{prefix_with_datetime}_{recording_filename_without_extension}'
            else:
                roi_filename_without_extension = self.roi.filepath.name.replace(self.roi.filepath.suffix, '')
                results_dir_name = f'{prefix_with_datetime}_{recording_filename_without_extension}_with_{roi_filename_without_extension}'
            self.results_dir_path = self.parent_dir_path.joinpath(results_dir_name)
            self.results_dir_path.mkdir()


    def _create_and_save_csv_result_files(self, filtered_squares: List[Square]) -> None:
        if len(filtered_squares) > 0:
            peak_results_per_square = [results.export_peak_results_df_from_square(square) for square in filtered_squares]
            df_all_peak_results = pd.concat(peak_results_per_square, ignore_index = True)
            max_peak_count_across_all_squares = df_all_peak_results.groupby('square coordinates [X / Y]').count()['peak frame index'].max()
            zfill_factor = int(np.log10(max_peak_count_across_all_squares)) + 1
            amplitude_and_delta_f_over_f_results_all_squares = []
            auc_results_all_squares = []
            for square_coords in df_all_peak_results['square coordinates [X / Y]'].unique():
                tmp_df_single_square = df_all_peak_results[df_all_peak_results['square coordinates [X / Y]'] == square_coords].copy()
                amplitude_and_delta_f_over_f_results_all_squares.append(results.create_single_square_amplitude_and_delta_f_over_f_results(tmp_df_single_square, zfill_factor))
                auc_results_all_squares.append(results.create_single_square_auc_results(tmp_df_single_square, zfill_factor))
            df_all_amplitude_and_delta_f_over_f_results = pd.concat(amplitude_and_delta_f_over_f_results_all_squares, ignore_index = True)
            df_all_auc_results = pd.concat(auc_results_all_squares, ignore_index = True)
            # Once all DataFrames are created successfully, write them to disk 
            df_all_peak_results.to_csv(self.results_dir_path.joinpath('all_peak_results.csv'), index = False)
            df_all_amplitude_and_delta_f_over_f_results.to_csv(self.results_dir_path.joinpath('Amplitude_and_dF_over_F_results.csv'), index = False)
            df_all_auc_results.to_csv(self.results_dir_path.joinpath('AUC_results.csv'), index = False)


    def _create_and_save_individual_traces_pdf_result_file(self, filtered_squares: List[Square], window_size: int) -> None:
            filepath = self.results_dir_path.joinpath('Individual_traces_with_identified_events.pdf')
            with PdfPages(filepath) as pdf:
                for indicate_activity in [True, False]:
                    roi = self.roi if self.roi_based == True else None
                    overview_fig, ax = results.plot_activity_overview(squares_with_sufficient_activity = filtered_squares,
                                                                      preview_image = self.recording.preview, 
                                                                      row_cropping_idx = self.row_cropping_idx, 
                                                                      col_cropping_idx = self.col_cropping_idx, 
                                                                      window_size = window_size, 
                                                                      indicate_activity = indicate_activity,
                                                                      roi = roi)
                    pdf.savefig(overview_fig)
                    plt.close()
                for square in filtered_squares:
                    fig = results.plot_intensity_trace_with_identified_peaks_for_individual_square(square)
                    pdf.savefig(fig)
                    plt.close()


import numpy as np
from scipy import signal
from pybaselines import Baseline
from collections import Counter
from shapely import Polygon

from typing import Optional, Tuple, Dict, List, Callable
from dataclasses import dataclass


class BaselineEstimatorFactory:
        
    @property
    def supported_baseline_estimation_methods(self) -> Dict[str, Callable]:
        supported_baseline_estimation_methods = {'asls': Baseline().asls,
                                                 'fabc': Baseline().fabc,
                                                 'psalsa': Baseline().psalsa,
                                                 'std_distribution': Baseline().std_distribution}
        return supported_baseline_estimation_methods

    def get_baseline_estimation_callable(self, algorithm_acronym: str) -> Callable:
        baseline_estimation_method = self.supported_baseline_estimation_methods[algorithm_acronym]
        return baseline_estimation_method



@dataclass
class Peak:

    frame_idx: int
    intensity: float
    amplitude: Optional[float]=None
    delta_f_over_f: Optional[float]=None
    has_neighboring_intersections: Optional[bool]=None
    frame_idxs_of_neighboring_intersections: Optional[Tuple[int, int]]=None
    area_under_curve: Optional[float]=None
    peak_type: Optional[str]=None



class Square:
    def __init__(self, grid_cell_label: Tuple[int, int], upper_left_corner_coords: Tuple[int, int], frames_zstack: np.ndarray) -> None:
        self.grid_row_label, self.grid_col_label = grid_cell_label
        self.upper_left_corner_coords = upper_left_corner_coords
        self.frames_zstack = frames_zstack
        self.center_coords = self._get_center_coords()
        self.as_polygon = self._create_square_as_polygon()
        self.peaks_count = 0

    
    def _get_center_coords(self) -> Tuple[int, int]:
        square_height = self.frames_zstack.shape[1]
        square_width = self.frames_zstack.shape[2]
        return (self.upper_left_corner_coords[0] + int(square_height/2), self.upper_left_corner_coords[1] + int(square_width/2))

    def _create_square_as_polygon(self) -> Polygon:
        square_rows = self.frames_zstack.shape[1]
        square_cols = self.frames_zstack.shape[2]
        upper_left_corner_row_coord = self.upper_left_corner_coords[0]
        upper_left_corner_col_coord = self.upper_left_corner_coords[1]
        all_corner_coords = [[upper_left_corner_row_coord, upper_left_corner_col_coord],
                             [upper_left_corner_row_coord, upper_left_corner_col_coord + square_cols],
                             [upper_left_corner_row_coord + square_cols, upper_left_corner_col_coord + square_cols],
                             [upper_left_corner_row_coord + square_cols, upper_left_corner_col_coord]]
        square_as_polygon = Polygon(all_corner_coords)
        assert square_as_polygon.is_valid, (
            f'Something went wrong when trying to create a Polygon for Square [{self.grid_col_label}/{self.self.grid_row_label}]!'
        )
        return square_as_polygon

    
    def compute_mean_intensity_timeseries(self, limit_analysis_to_frame_interval: bool, start_frame_idx: int, end_frame_idx: int) -> None:
        if limit_analysis_to_frame_interval == True:
            self.mean_intensity_over_time = np.mean(self.frames_zstack[start_frame_idx:end_frame_idx], axis = (1,2,3))
        else:
            self.mean_intensity_over_time = np.mean(self.frames_zstack, axis = (1,2,3))


    def detect_peaks(self, signal_to_noise_ratio: float, octaves_ridge_needs_to_spann: float, noise_window_size: int) -> None:
        widths = np.logspace(np.log10(1), np.log10(self.mean_intensity_over_time.shape[0]), 100)
        min_length = octaves_ridge_needs_to_spann / np.log2(widths[1] / widths[0])
        n_padded_frames = int(np.median(widths)) + 1
        signal_padded_with_reflection = np.pad(self.mean_intensity_over_time, n_padded_frames, 'reflect')
        frame_idxs_of_peaks_in_padded_signal = signal.find_peaks_cwt(vector = signal_padded_with_reflection, 
                                                         wavelet = signal.ricker, 
                                                         widths = widths, 
                                                         min_length = min_length,
                                                         max_distances = widths / 4, # default
                                                         gap_thresh = 0.0,
                                                         noise_perc = 5, # default: 10
                                                         min_snr = signal_to_noise_ratio,
                                                         window_size = noise_window_size # window size to calculate noise is very narrow (lowest point = noise)
                                                        )
        frame_idxs_of_peaks_in_padded_signal = frame_idxs_of_peaks_in_padded_signal[((frame_idxs_of_peaks_in_padded_signal >= n_padded_frames) & 
                                                                                     (frame_idxs_of_peaks_in_padded_signal < self.mean_intensity_over_time.shape[0] + n_padded_frames))]
        self.frame_idxs_of_peaks = frame_idxs_of_peaks_in_padded_signal - n_padded_frames
        self.peaks = {}
        for peak_frame_idx in self.frame_idxs_of_peaks:
            self.peaks[peak_frame_idx] = Peak(frame_idx = peak_frame_idx, intensity = self.mean_intensity_over_time[peak_frame_idx]) 
        self.peaks_count = self.frame_idxs_of_peaks.shape[0]


    def estimate_baseline(self, algorithm_acronym: str) -> None:
        baseline_estimation_method = BaselineEstimatorFactory().get_baseline_estimation_callable(algorithm_acronym)
        self.baseline = baseline_estimation_method(data = self.mean_intensity_over_time)[0]


    def compute_area_under_curve(self) -> None:
        self._get_unique_frame_idxs_of_intersections_between_signal_and_baseline()
        self._add_information_about_neighboring_intersections_to_peaks()
        area_under_curve_classification = {'peaks_with_auc': [], 'all_intersection_frame_idxs_pairs': []}
        for peak_frame_idx, peak in self.peaks.items():
            if peak.has_neighboring_intersections == True:
                idx_before_peak, idx_after_peak = peak.frame_idxs_of_neighboring_intersections
                peak.area_under_curve = np.trapz(self.mean_intensity_over_time[idx_before_peak:idx_after_peak + 1] - self.baseline[idx_before_peak:idx_after_peak + 1])
                area_under_curve_classification['peaks_with_auc'].append(peak)
                area_under_curve_classification['all_intersection_frame_idxs_pairs'].append(peak.frame_idxs_of_neighboring_intersections)
        self._classify_area_under_curve_types(area_under_curve_classification)
                                                                                    

    def _get_unique_frame_idxs_of_intersections_between_signal_and_baseline(self) -> None:
        quick_estimate_of_intersection_frame_idxs = np.argwhere(np.diff(np.sign(self.mean_intensity_over_time - self.baseline))).flatten()
        intersection_frame_idxs = np.asarray([self._improve_intersection_frame_idx_estimation_by_interpolation(idx) for idx in quick_estimate_of_intersection_frame_idxs])
        self.unique_intersection_frame_idxs = np.unique(intersection_frame_idxs)


    def _add_information_about_neighboring_intersections_to_peaks(self) -> None:
        for peak_frame_idx, peak in self.peaks.items():
            if (peak_frame_idx > self.unique_intersection_frame_idxs[0]) & (peak_frame_idx < self.unique_intersection_frame_idxs[-1]):
                peak.has_neighboring_intersections = True
                idx_pre_peak = self.unique_intersection_frame_idxs[self.unique_intersection_frame_idxs < peak_frame_idx][-1]
                idx_post_peak = self.unique_intersection_frame_idxs[self.unique_intersection_frame_idxs > peak_frame_idx][0]
                peak.frame_idxs_of_neighboring_intersections = (idx_pre_peak, idx_post_peak)
            else:
                peak.has_neighboring_intersections = False


    def _improve_intersection_frame_idx_estimation_by_interpolation(self, idx_frame_0: int) -> int:
        """
        Designed to resolve the bias of the quick estimation of intersection points, which will always 
        return the first index of two frames between which an intersection was determined. This is done 
        by interpolating the data (for both signal & baseline) to a sub-frame resolution between the 
        previously identified intersection frame index, and the following frame index - as the intersection 
        might actually happen closer to this following frame. If interpolation estimates the intersection 
        precisely in the middle between the two frames, the later frame is returned (0.5 is rounded up).
        """
        idx_frame_1 = idx_frame_0 + 1
        num_interpolated_steps = 7
        # interpolate signal & baseline to sub-frame resolution:
        interpolated_signal_intensities = np.linspace(self.mean_intensity_over_time[idx_frame_0], self.mean_intensity_over_time[idx_frame_1], num = num_interpolated_steps)
        interpolated_baseline = np.linspace(self.baseline[idx_frame_0], self.baseline[idx_frame_1], num = num_interpolated_steps)
        # identify whether frame_idx_0 or frame_idx_1 is closer to interpolated intersection point
        signed_differences = np.sign(interpolated_signal_intensities - interpolated_baseline)
        if 0 in signed_differences: #intersection exactly at one or multiple interpolated index
            intersection_idx_in_interpolation = np.argwhere(signed_differences == 0).flatten()[0]
        else: 
            results_for_intersection_idxs = np.argwhere(np.diff(signed_differences)).flatten()
            assert results_for_intersection_idxs.size != 0, ('get_improved_intersection_idx_estimation_by_interpolation() expected an intersection between frames '
                                                             f'{idx_frame_0} and {idx_frame_1} in the provided arrays, but none were found!')
            intersection_idx_in_interpolation = results_for_intersection_idxs[0]
        # Based on interpolation, select whether intersection is closer to frame_idx_0 or frame_idx_1
        if intersection_idx_in_interpolation < np.median(np.arange(0, num_interpolated_steps, 1)):
            interpolation_evaluated_intersection_frame_idx = idx_frame_0
        else:
            interpolation_evaluated_intersection_frame_idx = idx_frame_1
        return interpolation_evaluated_intersection_frame_idx
    

    def _classify_area_under_curve_types(self, data_for_auc_classification: Dict[str, List]) -> None:
        if len(data_for_auc_classification['all_intersection_frame_idxs_pairs']) != len(set(data_for_auc_classification['all_intersection_frame_idxs_pairs'])):
            counter = Counter(data_for_auc_classification['all_intersection_frame_idxs_pairs'])
            reoccuring_intersection_frame_idxs = [pair_of_intersection_frame_idxs for pair_of_intersection_frame_idxs, count in counter.items() if count > 1]
        else:
            reoccuring_intersection_frame_idxs = []
        for peak in self.peaks.values():
            if peak in data_for_auc_classification['peaks_with_auc']:
                if peak.frame_idxs_of_neighboring_intersections in reoccuring_intersection_frame_idxs:
                    peak.peak_type = 'clustered'
                else:
                    peak.peak_type = 'singular'
            else:
                peak.peak_type = 'isolated'

    
    def compute_amplitude_and_delta_f_over_f(self):
        for peak in self.peaks.values():
            peak.amplitude = self.mean_intensity_over_time[peak.frame_idx] - self.baseline[peak.frame_idx]
            peak.delta_f_over_f = peak.amplitude / self.baseline[peak.frame_idx]










import matplotlib.pyplot as plt
from matplotlib.transforms import Bbox
import numpy as np
import pandas as pd

from typing import List, Tuple, Optional
from pathlib import Path
from matplotlib.text import Text
from matplotlib.figure import Figure
from matplotlib.axes._axes import Axes 

from .analysis import Square
from .input import ROI


def _get_text_bounding_box_size_in_data_dimensions(text: Text, fig: Figure, ax: Axes) -> Tuple[float, float]:
    renderer = fig.canvas.get_renderer()
    text_bbox_raw_dimensions = text.get_window_extent(renderer=renderer)
    text_bbox_data_dimensions = Bbox(ax.transData.inverted().transform(text_bbox_raw_dimensions))
    return np.abs(text_bbox_data_dimensions.width), np.abs(text_bbox_data_dimensions.height)


def _iteratively_decrease_fontsize_to_fit_text_in_squares(text: Text, max_size: float, fig: Figure, ax: Axes) -> None:
    text.set_fontsize(text.get_fontsize()-1)
    text_width, text_height = _get_text_bounding_box_size_in_data_dimensions(text, fig, ax)
    if (text_width > max_size) or (text_height > max_size): 
        _iteratively_decrease_fontsize_to_fit_text_in_squares(text, max_size, fig, ax)


def _get_adjusted_fontsize(preview_image: np.ndarray, window_size: int, max_peak_count: int, default_fontsize: int=25) -> float:
    max_text_size = window_size*0.75
    tmp_fig, tmp_ax = plt.subplots()
    tmp_ax.imshow(preview_image)
    sample_coord = window_size + 0.5 * window_size
    max_width_text_from_number = '4'*len(str(max_peak_count))
    tmp_text = tmp_ax.text(sample_coord, sample_coord, max_width_text_from_number, fontsize = default_fontsize)
    text_width, text_height = _get_text_bounding_box_size_in_data_dimensions(tmp_text, tmp_fig, tmp_ax)
    if (text_width > max_text_size) or (text_height > max_text_size):
        _iteratively_decrease_fontsize_to_fit_text_in_squares(tmp_text, max_text_size, tmp_fig, tmp_ax)
    adjusted_fontsize = tmp_text.get_fontsize()
    plt.close(tmp_fig)
    return adjusted_fontsize


def plot_window_size_preview(preview_image: np.ndarray, window_size: int, row_cropping_idx: int, col_cropping_idx: int, roi: Optional[ROI]=None) -> Tuple[Figure, Axes]:
    fig, ax = plt.subplots()
    ax.imshow(preview_image, cmap="gray", vmin = np.percentile(preview_image, 2.5), vmax = np.percentile(preview_image, 97.5))
    ax.grid(color = 'gray', linestyle = '--', linewidth = 1)
    if roi == None:
        plt.hlines([0, row_cropping_idx], xmin=0, xmax=col_cropping_idx, color = 'cyan', linewidth = 2)
        plt.vlines([0, col_cropping_idx], ymin=0, ymax=row_cropping_idx, color = 'cyan', linewidth = 2)
    else:
        roi_boundary_coords_for_plotting = np.asarray(roi.boundary_coords)
        plt.plot(roi_boundary_coords_for_plotting[:, 1], roi_boundary_coords_for_plotting[:, 0], c = 'cyan', linewidth = 2)
    ax.set_xticks(np.arange(0, preview_image.shape[1], window_size), labels = [])
    ax.set_xticks(np.arange(window_size/2, col_cropping_idx + window_size/2, window_size), 
                  labels = np.arange(1, col_cropping_idx/window_size + 1, 1, dtype='int'), minor = True)
    ax.xaxis.set_label_text('X')
    ax.set_yticks(np.arange(0,  preview_image.shape[0], window_size), labels = [])
    ax.set_yticks(np.arange(window_size/2, row_cropping_idx + window_size/2, window_size), 
                  labels = np.arange(1, row_cropping_idx/window_size + 1, 1, dtype='int'), minor = True)
    ax.yaxis.set_label_text('Y')
    ax.tick_params(bottom = False, left = False)
    ax.set_title(f'Preview of grid for window size: {window_size}')
    return fig, ax



def plot_activity_overview(squares_with_sufficient_activity: List[Square], 
                           preview_image: np.ndarray, 
                           row_cropping_idx: int, 
                           col_cropping_idx: int, 
                           window_size: int, 
                           indicate_activity: bool=False,
                           roi: Optional[ROI]=None
                          ) -> Tuple[Figure, Axes]:
    all_peak_counts = [square.peaks_count for square in squares_with_sufficient_activity]
    if len(all_peak_counts) > 0:
        max_peak_count = max(all_peak_counts)
    else:
        max_peak_count = 0
    peak_text_fontsize = _get_adjusted_fontsize(preview_image, window_size, max_peak_count)
    fig, ax = plt.subplots()
    ax.imshow(preview_image, cmap="gray", vmin = np.percentile(preview_image, 2.5), vmax = np.percentile(preview_image, 97.5))
    if indicate_activity == True:
        for square in squares_with_sufficient_activity:
            ax.text(square.center_coords[1], 
                    square.center_coords[0], 
                    square.peaks_count, 
                    color = 'magenta', 
                    horizontalalignment='center', 
                    verticalalignment = 'center', 
                    fontsize = peak_text_fontsize)
    ax.grid(color = 'gray', linestyle = '--', linewidth = 1)
    if roi == None:
        plt.hlines([0, row_cropping_idx], xmin=0, xmax=col_cropping_idx, color = 'cyan', linewidth = 2)
        plt.vlines([0, col_cropping_idx], ymin=0, ymax=row_cropping_idx, color = 'cyan', linewidth = 2)
    else:
        roi_boundary_coords_for_plotting = np.asarray(roi.boundary_coords)
        plt.plot(roi_boundary_coords_for_plotting[:, 1], roi_boundary_coords_for_plotting[:, 0], c = 'cyan', linewidth = 2)
    ax.set_xticks(np.arange(0, preview_image.shape[1], window_size), labels = [])
    ax.set_xticks(np.arange(window_size/2, col_cropping_idx + window_size/2, window_size), 
                  labels = np.arange(1, col_cropping_idx/window_size + 1, 1, dtype='int'), 
                  minor = True, 
                  fontsize = min(12, peak_text_fontsize))
    ax.xaxis.set_label_text('X')
    ax.set_yticks(np.arange(0,  preview_image.shape[0], window_size), labels = [])
    ax.set_yticks(np.arange(window_size/2, row_cropping_idx + window_size/2, window_size), 
                  labels = np.arange(1, row_cropping_idx/window_size + 1, 1, dtype='int'), 
                  minor = True, 
                  fontsize = min(12, peak_text_fontsize))
    ax.yaxis.set_label_text('Y')
    ax.tick_params(bottom = False, left = False)
    ax.set_title(f'Total activity: {np.sum(all_peak_counts)}')
    return fig, ax


def plot_intensity_trace_with_identified_peaks_for_individual_square(square: Square) -> Figure:
    fig = plt.figure(figsize = (9, 2.67), facecolor = 'white')
    plt.plot(square.mean_intensity_over_time, c = 'gray')
    if hasattr(square, 'baseline'):
        plt.plot(square.baseline, c = 'cyan')
    for peak in square.peaks.values():
        if peak.has_neighboring_intersections == True:
            plt.plot(peak.frame_idx, peak.intensity, 'mo')
            start_idx = peak.frame_idxs_of_neighboring_intersections[0] - 1
            if start_idx < 0:
                start_idx = 0
            end_idx = peak.frame_idxs_of_neighboring_intersections[1] + 1
            plt.fill_between(np.arange(start_idx, end_idx, 1), 
                             square.mean_intensity_over_time[start_idx : end_idx], 
                             square.baseline[start_idx : end_idx], 
                             where = square.mean_intensity_over_time[start_idx : end_idx] > square.baseline[start_idx : end_idx], 
                             interpolate = True, 
                             color='yellow',
                             alpha = 0.6)
        else:
            plt.plot(peak.frame_idx, peak.intensity, 'ko')
    plt.title(f'Graph: [{square.grid_col_label} / {square.grid_row_label}]     Total Activity: {square.peaks_count}')
    plt.tight_layout()
    return fig



def export_peak_results_df_from_square(square: Square) -> pd.DataFrame:
    all_peaks = [peak for peak in square.peaks.values()]
    df_all_peak_results_one_square = pd.DataFrame(all_peaks)
    df_all_peak_results_one_square.drop(['has_neighboring_intersections', 'frame_idxs_of_neighboring_intersections'], axis = 'columns', inplace = True)
    df_all_peak_results_one_square.columns = ['peak frame index', 'peak bit value', 'peak amplitude', 'peak dF/F',  'peak AUC', 'peak classification']
    df_all_peak_results_one_square.insert(loc = 0, column = 'square coordinates [X / Y]', value = f'[{square.grid_col_label} / {square.grid_row_label}]')
    return df_all_peak_results_one_square



def create_single_square_amplitude_and_delta_f_over_f_results(df_all_results_single_square: pd.DataFrame, zfill_factor: int) -> pd.DataFrame:
    rearranged_data = {'square coordinates [X / Y]': [df_all_results_single_square['square coordinates [X / Y]'].iloc[0]],
                       'total peak count': [df_all_results_single_square.shape[0]]}
    for i in range(df_all_results_single_square.shape[0]):
        peak_idx = str(i + 1)
        peak_idx_suffix = peak_idx.zfill(zfill_factor)
        rearranged_data[f'frame index peak #{peak_idx_suffix}'] = [df_all_results_single_square.iloc[i]['peak frame index']]
        rearranged_data[f'amplitude peak #{peak_idx_suffix}'] = [df_all_results_single_square.iloc[i]['peak amplitude']]
        rearranged_data[f'dF/F peak #{peak_idx_suffix}'] = [df_all_results_single_square.iloc[i]['peak dF/F']]
    return pd.DataFrame(rearranged_data)
    

    
def create_single_square_auc_results(df_all_results_single_square: pd.DataFrame, zfill_factor: int) -> pd.DataFrame:
    rearranged_data = {'square coordinates [X / Y]': [df_all_results_single_square['square coordinates [X / Y]'].iloc[0]],
                       'total count all peaks': [df_all_results_single_square.shape[0]],
                       'total count "singular" peaks': [df_all_results_single_square['peak classification'].str.count('singular').sum()],
                       'total count "clustered" peaks': [df_all_results_single_square['peak classification'].str.count('clustered').sum()],
                       'total count "isolated" peaks': [df_all_results_single_square['peak classification'].str.count('isolated').sum()]}
    for i in range(df_all_results_single_square.shape[0]):
        peak_idx = str(i + 1)
        peak_idx_suffix = peak_idx.zfill(zfill_factor)
        rearranged_data[f'frame index peak #{peak_idx_suffix}'] = [df_all_results_single_square.iloc[i]['peak frame index']]
        rearranged_data[f'AUC peak #{peak_idx_suffix}'] = [df_all_results_single_square.iloc[i]['peak AUC']]
        rearranged_data[f'classification peak #{peak_idx_suffix}'] = [df_all_results_single_square.iloc[i]['peak classification']]
    return pd.DataFrame(rearranged_data)




from pathlib import Path
import json
from datetime import datetime, timezone
import ipywidgets as w
import matplotlib.pyplot as plt
import inspect
import multiprocessing

from .processing import AnalysisJob
from .input import RecordingLoaderFactory, ROILoaderFactory, RecLoaderROILoaderCombinator, get_filepaths_with_supported_extension_in_dirpath

from typing import Tuple, Dict, Callable, Optional, Any
from matplotlib.figure import Figure
from matplotlib.axes._axes import Axes 



class Model:

    def __init__(self) -> None:
        self.num_processes = multiprocessing.cpu_count()
        self.analysis_job_queue = []
        self.logs = []
        self.gui_enabled = False


    def setup_connection_to_update_infos_in_view(self, update_infos: Callable) -> None:
        self.callback_view_update_infos = update_infos
        self._check_if_gui_setup_is_completed()


    def setup_connection_to_display_results(self, show_output_screen: Callable, output: w.Output, pixel_conversion: float) -> None:
        self.callback_view_show_output_screen = show_output_screen
        self.view_output = output
        self.pixel_conversion = pixel_conversion
        self._check_if_gui_setup_is_completed()


    def _check_if_gui_setup_is_completed(self) -> None:
        checklist_expected_attributes_with_type = {'callback_view_update_infos': Callable,
                                                   'callback_view_show_output_screen': Callable,
                                                   'view_output': w.Output,
                                                   'pixel_conversion': float}
        confirmed_attributes = []
        failed_attributes = []
        for attribute_name, expected_type in checklist_expected_attributes_with_type.items():
            if hasattr(self, attribute_name) == True:
                attribute = getattr(self, attribute_name)
                if type(attribute) == expected_type:
                    confirmed_attributes.append(True)
                elif isinstance(attribute, expected_type):
                    confirmed_attributes.append(True)
                else:
                    failed_attributes.append(attribute_name)
            else:
                failed_attributes.append(attribute_name)
        if len(confirmed_attributes) == len(checklist_expected_attributes_with_type.keys()):
            self.gui_enabled = True
            self.add_info_to_logs('NA3 GUI initialization completed. Start by selecting and loading source data.', True)
        else:
            self.gui_enabled = False
            for attribute_name in failed_attributes:
                self.add_info_to_logs(f'Setup of {attribute_name} missing before GUI connection can be enabled.')


    def add_info_to_logs(self, message: str, display_in_gui: bool=False, progress_in_percent: Optional[float]=None) -> None:
        time_prefix_in_utc = datetime.now(timezone.utc).strftime('%d-%m-%y %H:%M:%S.%f')
        self.logs.append(f'{time_prefix_in_utc} (UTC): {message}')
        if (display_in_gui == True) and (self.gui_enabled == True): 
            self.callback_view_update_infos(message, progress_in_percent)

    
    def create_analysis_jobs(self, configs: Dict[str, Any]) -> None:
        self._ensure_data_from_previous_jobs_was_removed()
        validated_configs = self._get_configs_required_for_specific_function(configs, self._assertion_for_create_analysis_jobs)
        self.add_info_to_logs('Basic configurations for data import validated. Starting creation of analysis job(s)...', True)
        if validated_configs['batch_mode'] == False:
            if validated_configs['roi_mode'] == False:
                self._add_new_recording_without_rois_to_analysis_job_queue(validated_configs['data_source_path'], validated_configs['batch_mode'])
            else: #'roi_mode' == True:
                self._add_new_recording_with_rois_to_analysis_job_queue(validated_configs['data_source_path'])
        else: #'batch_mode' == True:
            all_subdirs = [subdir for subdir in validated_configs['data_source_path'].iterdir() if subdir.is_dir() and (subdir.name.startswith('.') == False)]
            all_subdirs.sort()
            total_step_count = len(all_subdirs)
            progress_step_size = 100 / total_step_count
            if validated_configs['roi_mode'] == False:
                for idx, subdir_path in enumerate(all_subdirs):
                    self._add_new_recording_without_rois_to_analysis_job_queue(subdir_path, validated_configs['batch_mode'])
                    self.add_info_to_logs(f'Job creation for {subdir_path} completed.', True, min((idx+1)*progress_step_size, 100.0))
            else: #'roi_mode' == True:
                for idx, subdir_path in enumerate(all_subdirs):
                    self._add_new_recording_with_rois_to_analysis_job_queue(subdir_path)
                    self.add_info_to_logs(f'Job creation(s) for {subdir_path} completed.', True, min((idx+1)*progress_step_size, 100.0))
        self.add_info_to_logs('All job creation(s) completed.', True, 100.0)
        

    def _ensure_data_from_previous_jobs_was_removed(self) -> None:
        if len(self.analysis_job_queue) > 0:
            self.add_info_to_logs('Loading of new source data. All previously created jobs & logs will be deleted.', True)
            self.analysis_job_queue = []
            self.logs = []
            self._check_if_gui_setup_is_completed()

    
    def _assertion_for_create_analysis_jobs(self, batch_mode: bool, roi_mode: bool, data_source_path: Path) -> None:
        # just a convenience function to use the existing config validation and filtering methods
        pass
        

    def _add_new_recording_without_rois_to_analysis_job_queue(self, filepath: Path, batch_mode: bool) -> None:
        rec_loader_factory = RecordingLoaderFactory()
        job_idx = len(self.analysis_job_queue)
        if batch_mode == True:
            self.add_info_to_logs(f'Looking for a valid recording file in {filepath}...', True)
            valid_filepaths = get_filepaths_with_supported_extension_in_dirpath(filepath, rec_loader_factory.all_supported_extensions, 1)
            if len(valid_filepaths) == 0:
                self.add_info_to_logs(f'Could not find any recording files of supported type at {filepath}!', True)
            elif len(valid_filepaths) >  1:
                too_many_files_message = (f'Found more than a single recording file of supported type at {filepath}, i.e.: {valid_filepaths}. '
                                          f'However, only a single file was expected. NA3 continues with {valid_filepaths[0]} and will ignore the other files.')
                self.add_info_to_logs(too_many_files_message, True)
                recording_filepath = valid_filepaths[0]
            else:
                recording_filepath = valid_filepaths[0]
        else:
            recording_filepath = filepath
        self.add_info_to_logs(f'Starting to create a job for: {recording_filepath}', True)
        try:
            recording_loader = rec_loader_factory.get_loader(recording_filepath)
        except NotImplementedError:
            message = (f'Job creation failed! The data you selected ("{recording_filepath}") is not a supported recording file type! '
                       f'Currently supported file types for recordings are: {rec_loader_factory.all_supported_extensions}.')
            if batch_mode == True:
                message += 'For your convenience, this job is skipped and NA3 continues with the next one instead.'
            self.add_info_to_logs(message, True)
        except Exception as e:
            message = 'Job creation failed due to an unexpected error: '
            all_unexpected_exceptions = [exception_message for exception_message in e.args]
            for unexpected_exception in all_unexpected_exceptions:
                message += f'"{unexpected_exception}"; '
            self.add_info_to_logs(message, True)
            self.add_info_to_logs('Continuing with next job, if available.', True)
        else: # no errors in try
            self.analysis_job_queue.append(AnalysisJob(self.num_processes, recording_loader))
            self.add_info_to_logs(f'Successfully created job for: {recording_filepath} at index #{job_idx}.', True)


    def _add_new_recording_with_rois_to_analysis_job_queue(self, dir_path) -> None:
        self.add_info_to_logs(f'Looking for a recording file and ROI files in {dir_path} and trying to create corresponding jobs...', True)
        rec_roi_loader = RecLoaderROILoaderCombinator(dir_path)
        try:
            recording_and_roi_loader_combos = rec_roi_loader.get_all_recording_and_roi_loader_combos()
        except NotImplementedError:
            rec_loader_factory = RecordingLoaderFactory()
            message = (f'Job creation failed! Could not find a supported recording file in "{dir_path}"! '
                       f'Currently supported file types for recordings are: {rec_loader_factory.all_supported_extensions}. '
                        'For your convenience, this job is skipped and NA3 continues with the next one instead.')
            self.add_info_to_logs(message, True)
        except Exception as e:
            message = 'Job creation failed due to an unexpected error: '
            all_unexpected_exceptions = [exception_message for exception_message in e.args]
            for unexpected_exception in all_unexpected_exceptions:
                message += f'"{unexpected_exception}"; '
            self.add_info_to_logs(message, True)
            self.add_info_to_logs('Continuing with next job, if available.', True)
        else: # no errors in try
            for recording_loader, roi_loader in recording_and_roi_loader_combos:
                job_idx = len(self.analysis_job_queue)
                self.analysis_job_queue.append(AnalysisJob(self.num_processes, recording_loader, roi_loader))
                self.add_info_to_logs(f'Successfully created job for: {recording_loader.filepath} with {roi_loader.filepath} at index #{job_idx}.', True)
    

    def run_analysis(self, configs: Dict[str, Any]) -> None:
        sample_job_to_validate_configs = self.analysis_job_queue[0]
        validated_configs_for_analysis = self._get_configs_required_for_specific_function(configs, sample_job_to_validate_configs.run_analysis)
        validated_configs_for_result_creation = self._get_configs_required_for_specific_function(configs, sample_job_to_validate_configs.create_results)
        self.add_info_to_logs('Configurations for Analysis Settings and Result Creation validated successfully.', True)
        self.add_info_to_logs(f'Analysis Settings are:')
        for key, value in validated_configs_for_analysis.items():
            self.add_info_to_logs(f'{key}: {str(value)}')     
        self.add_info_to_logs('Starting analysis...', True)
        total_step_count = len(self.analysis_job_queue)
        progress_step_size = 100 / total_step_count
        for job_idx, analysis_job in enumerate(self.analysis_job_queue):
            self.add_info_to_logs(f'Starting to process analysis job with index #{job_idx}.')
            analysis_job.run_analysis(**validated_configs_for_analysis)
            self.add_info_to_logs(f'Analysis successfully completed. Continue with creation of results.. ')
            analysis_job.create_results(**validated_configs_for_result_creation)
            self.add_info_to_logs(f'Results successfully created at: {analysis_job.results_dir_path}')
            if self.gui_enabled == True:
                self.callback_view_show_output_screen()
                with self.view_output:
                    overview_results_fig, overview_results_ax = analysis_job.overview_results
                    overview_results_fig.set_figheight(400 * self.pixel_conversion)
                    overview_results_fig.tight_layout()
                    plt.show(overview_results_fig)
            self._save_user_settings_as_json(configs, analysis_job)
            self.add_info_to_logs(f'Successfully finished processing of analysis job with index #{job_idx}. {job_idx + 1} of {total_step_count} total job(s) completed.', 
                                  True,
                                  min((job_idx+1)*progress_step_size, 100.0))
            self._save_logs_at_current_moment(analysis_job.results_dir_path)
        self.add_info_to_logs('Updating all log files to contain all logs as final step. All valid logs files will end with this message.')
        for job_idx, analysis_job in enumerate(self.analysis_job_queue):
            self._save_logs_at_current_moment(analysis_job.results_dir_path)
            

    def _get_configs_required_for_specific_function(self, all_configs: Dict[str, Any], function_to_execute: Callable) -> Dict[str, Any]:
        filtered_and_validated_configs = {}
        for expected_parameter_name in inspect.signature(function_to_execute).parameters:
            self._validate_individual_config_value_against_function_type_hints(all_configs, function_to_execute, expected_parameter_name)
            filtered_and_validated_configs[expected_parameter_name] = all_configs[expected_parameter_name]
        return filtered_and_validated_configs


    def _validate_individual_config_value_against_function_type_hints(self, all_configs: Dict[str, Any], function_to_execute: Callable, expected_parameter_name: str) -> None:
        assert expected_parameter_name in all_configs.keys(), (
            f'{function_to_execute.__name__} requires the parameter "{expected_parameter_name}", '
            f'which is not included in the configs ({list(all_configs.keys())})'
        )
        value_in_configs = all_configs[expected_parameter_name]
        expected_parameter_type = inspect.signature(function_to_execute).parameters[expected_parameter_name].annotation
        if expected_parameter_type == Path:
            assert isinstance(value_in_configs, Path), (
                f'{function_to_execute.__name__} requires the parameter "{expected_parameter_name}" to be a '
                f'pathlib.Path object. However, {value_in_configs}, which is of type '
                f'{type(value_in_configs)}, was passed.'
            )
        else:
            assert expected_parameter_type == type(value_in_configs), (
                f'{function_to_execute.__name__} requires the parameter "{expected_parameter_name}" '
                f'to be of type {expected_parameter_type}. However, {value_in_configs}, which '
                f'is of type {type(value_in_configs)}, was passed.'
            )


    def _save_logs_at_current_moment(self, results_dir_path: Path) -> None:
        filepath = results_dir_path.joinpath('logs.txt')
        with open(filepath , 'w+') as logs_file:
            for log_message in self.logs:
                logs_file.write(f'{log_message}\n')


    def _save_user_settings_as_json(self, configs: Dict[str, Any], analysis_job: AnalysisJob) -> None:
        filepath = analysis_job.results_dir_path.joinpath('user_settings.json')
        configs['recording_filepath'] = analysis_job.recording.filepath
        if analysis_job.roi_based == True:
            configs['roi_filepath'] = analysis_job.roi.filepath
        else:
            configs['roi_filepath'] = None
        configs_preformatted_for_json = {}
        for key, value in configs.items():
            if isinstance(value, Path):
                configs_preformatted_for_json[key] = value.as_posix()
            else:
                configs_preformatted_for_json[key] = value
        with open(filepath, 'w+') as user_settings_json: 
            json.dump(configs_preformatted_for_json, user_settings_json)


    def preview_window_size(self, configs: Dict[str, Any]) -> Tuple[Figure, Axes]:
        job_for_preview = self.analysis_job_queue[0]
        validated_configs_for_preview = self._get_configs_required_for_specific_function(configs, job_for_preview.preview_window_size)
        preview_fig, preview_ax = job_for_preview.preview_window_size(**validated_configs_for_preview)
        return preview_fig, preview_ax




from .model import Model
from .view import WidgetsInterface
from PIL import Image
import matplotlib.pyplot as plt
from pathlib import Path
import numpy as np
from IPython.display import display

from typing import Any, Callable, Dict

class App:

    def __init__(self):
        self.model = Model()
        self.view = WidgetsInterface()
        self.pixel_conversion = 1/plt.rcParams['figure.dpi']
        self._setup_interaction_between_model_and_view()


    def _setup_interaction_between_model_and_view(self) -> None:
        self._bind_buttons_of_view_to_functions_of_model()
        self.model.setup_connection_to_update_infos_in_view(self.view.update_infos)
        self.model.setup_connection_to_display_results(self.view.main_screen.show_output_screen, self.view.main_screen.output, self.pixel_conversion)


    def _bind_buttons_of_view_to_functions_of_model(self) -> None:
        self.view.source_data_panel.load_source_data_button.on_click(self._load_data_button_clicked)
        self.view.analysis_settings_panel.run_analysis_button.on_click(self._run_button_clicked)
        self.view.analysis_settings_panel.preview_window_size_button.on_click(self._preview_window_size_button_clicked)

    
    def launch(self) -> None:
        display(self.view.widget)


    def _load_data_button_clicked(self, change) -> None:
        user_settings = self.view.export_user_settings()
        self.model.create_analysis_jobs(user_settings)
        if len(self.model.analysis_job_queue) < 1:
            self.model.add_info_to_logs('Failed to create any analysis job(s). Please inspect logs for more details!', True)
            self.view.user_info_panel.progress_bar.bar_style = 'danger'
        else:
            representative_job = self.model.analysis_job_queue[0]
            representative_job.load_data_into_memory()
            self.view.adjust_widgets_to_loaded_data(total_frames = representative_job.recording.zstack.shape[0])
            self.view.main_screen.show_output_screen()
            with self.view.main_screen.output:
                fig = plt.figure(figsize = (600*self.pixel_conversion, 400*self.pixel_conversion))
                if representative_job.roi_based == True:
                    roi_boundary_coords = np.asarray(representative_job.roi.boundary_coords)
                    plt.plot(roi_boundary_coords[:, 1], roi_boundary_coords[:, 0], c = 'cyan', linewidth = 2)
                    plt.imshow(representative_job.recording.preview, cmap = 'gray')
                else:
                    plt.imshow(representative_job.recording.preview, cmap = 'gray')
                plt.tight_layout()
                plt.show()
            self.model.add_info_to_logs(f'Data import completed! {len(self.model.analysis_job_queue)} job(s) in queue.', True, 100.0)
            self.view.enable_analysis()


    def _run_button_clicked(self, change) -> None:
        self.view.enable_analysis(False)
        user_settings = self.view.export_user_settings()
        self.model.run_analysis(user_settings)
        self.model.add_info_to_logs(f'Processing of all jobs completed! Feel free to load more data & continue analyzing!', True, 100.0)
        self.view.enable_analysis(True)


    def _preview_window_size_button_clicked(self, change) -> None:
        self.view.main_screen.show_output_screen()
        with self.view.main_screen.output:
            user_settings = self.view.export_user_settings()
            preview_fig, preview_ax = self.model.preview_window_size(user_settings)
            preview_fig.set_figheight(400 * self.pixel_conversion)
            preview_fig.tight_layout()
            plt.show(preview_fig)


def open_gui() -> None:
    na3 = App()
    return na3.launch()
